@@ -9,13 +9,13 @@ use pyo3::{
     basic::CompareOp,
     exceptions::{PyException, PyNotImplementedError, PyTypeError},
     prelude::*,
-    types::{PyDict, PyIterator},
+    types::{PyDict, PyIterator, PyTuple},
 };
 use rosu_pp::{
     beatmap::BeatmapAttributes, catch::CatchPerformanceAttributes,
     mania::ManiaPerformanceAttributes, osu::OsuPerformanceAttributes,
-    taiko::TaikoPerformanceAttributes, AnyPP, Beatmap, BeatmapExt, GameMode as RosuGameMode,
-    PerformanceAttributes,
+    taiko::TaikoPerformanceAttributes, AnyPP, Beatmap, BeatmapExt, GameMode, PerformanceAttributes,
+    Strains as RosuStrains,
 };
 
 #[pyclass]
@@ -88,7 +88,7 @@ impl Calculator {
         self.0.od = od;
     }
 
-    fn calculate(&mut self, py: Python, obj: &PyAny) -> PyResult<Vec<CalculateResult>> {
+    fn calculate(&mut self, obj: &PyAny) -> PyResult<Vec<CalculateResult>> {
         match obj.extract::<ScoreParams>() {
             Ok(params) => {
                 let mods = params.mods;
@@ -103,7 +103,7 @@ impl Calculator {
             Err(_) => {
                 let mut mod_diffs = HashMap::new();
 
-                PyIterator::from_object(py, obj)
+                PyIterator::from_object(obj.py(), obj)
                     .map_err(|_| {
                         let py_type = obj.get_type().name().unwrap_or("<unknown type>");
 
@@ -146,25 +146,103 @@ impl Calculator {
             }
         }
     }
+
+    #[args(args = "*")]
+    fn strains(&mut self, args: &PyTuple) -> PyResult<Strains> {
+        let mods = if let Ok(obj) = args.get_item(0) {
+            if let Ok(mods) = obj.extract::<u32>() {
+                mods
+            } else {
+                let py_type = obj.get_type().name().unwrap_or("<unknown type>");
+                let err = format!("got '{}'; expected 'int'", py_type);
+
+                return Err(PyTypeError::new_err(err));
+            }
+        } else {
+            0
+        };
+
+        Ok(self.0.strains(mods).into())
+    }
 }
 
 #[pyclass]
-#[derive(Copy, Clone, PartialEq)]
-#[repr(u8)]
-enum GameMode {
-    Osu,
-    Taiko,
-    Catch,
-    Mania,
+#[derive(Clone, Default, PartialEq)]
+#[allow(non_snake_case)]
+struct Strains {
+    #[pyo3(get, set)]
+    sectionLength: f64,
+
+    #[pyo3(get, set)]
+    color: Option<Vec<f64>>,
+    #[pyo3(get, set)]
+    rhythm: Option<Vec<f64>>,
+    #[pyo3(get, set)]
+    staminaLeft: Option<Vec<f64>>,
+    #[pyo3(get, set)]
+    staminaRight: Option<Vec<f64>>,
+
+    #[pyo3(get, set)]
+    aim: Option<Vec<f64>>,
+    #[pyo3(get, set)]
+    aimNoSliders: Option<Vec<f64>>,
+    #[pyo3(get, set)]
+    speed: Option<Vec<f64>>,
+    #[pyo3(get, set)]
+    flashlight: Option<Vec<f64>>,
+
+    #[pyo3(get, set)]
+    strains: Option<Vec<f64>>,
+
+    #[pyo3(get, set)]
+    movement: Option<Vec<f64>>,
 }
 
-impl From<GameMode> for RosuGameMode {
-    fn from(mode: GameMode) -> Self {
-        match mode {
-            GameMode::Osu => RosuGameMode::STD,
-            GameMode::Taiko => RosuGameMode::TKO,
-            GameMode::Catch => RosuGameMode::CTB,
-            GameMode::Mania => RosuGameMode::MNA,
+#[pymethods]
+impl Strains {
+    fn __richcmp__(&self, other: &PyAny, op: CompareOp) -> PyResult<bool> {
+        match (other.extract::<Self>(), op) {
+            (Ok(ref other), CompareOp::Eq) => Ok(self == other),
+            (Ok(ref other), CompareOp::Ne) => Ok(self != other),
+            _ => Err(PyNotImplementedError::new_err("")),
+        }
+    }
+
+    fn __repr__(&self) -> PyResult<String> {
+        Ok(self.to_string())
+    }
+}
+
+impl From<RosuStrains> for Strains {
+    #[inline]
+    fn from(strains: RosuStrains) -> Self {
+        match strains {
+            RosuStrains::Catch(strains) => Self {
+                sectionLength: strains.section_len,
+                movement: Some(strains.movement),
+                ..Default::default()
+            },
+            RosuStrains::Mania(strains) => Self {
+                sectionLength: strains.section_len,
+                strains: Some(strains.strains),
+                ..Default::default()
+            },
+            RosuStrains::Osu(strains) => Self {
+                sectionLength: strains.section_len,
+                aim: Some(strains.aim),
+                aimNoSliders: Some(strains.aim_no_sliders),
+                speed: Some(strains.speed),
+                flashlight: Some(strains.flashlight),
+                ..Default::default()
+            },
+            RosuStrains::Taiko(strains) => Self {
+                sectionLength: strains.section_len,
+                color: Some(strains.color),
+                rhythm: Some(strains.rhythm),
+                staminaLeft: Some(strains.stamina_left),
+                staminaRight: Some(strains.stamina_right),
+                ..Default::default()
+            },
         }
     }
 }
@@ -173,7 +251,7 @@ impl From<GameMode> for RosuGameMode {
 #[derive(Clone, Default, PartialEq)]
 struct ScoreParams {
     #[pyo3(get, set)]
-    mode: Option<GameMode>,
+    mode: Option<u8>,
     #[pyo3(get, set)]
     mods: u32,
     #[pyo3(get, set)]
@@ -424,8 +502,16 @@ impl ScoreParams {
             clock_rate,
         } = self;
 
-        if let Some(mode) = mode {
-            calculator = calculator.mode(mode.into());
+        if let Some(mode @ 0..=3) = mode {
+            let mode = match mode {
+                0 => GameMode::Osu,
+                1 => GameMode::Taiko,
+                2 => GameMode::Catch,
+                3 => GameMode::Mania,
+                _ => unreachable!(),
+            };
+
+            calculator = calculator.mode(mode);
         }
 
         if let Some(n300) = n300 {
@@ -563,6 +649,40 @@ impl ScoreParams {
     }
 }
 
+impl Display for Strains {
+    fn fmt(&self, f: &mut Formatter<'_>) -> FmtResult {
+        let mut s = f.debug_struct("Strains");
+
+        s.field("sectionLength", &self.sectionLength);
+
+        macro_rules! display_field {
+            ($self:ident, $s:ident: $($field:ident,)*) => {
+                $(
+                    if let Some(ref field) = $self.$field {
+                        $s.field(stringify!($field), field);
+                    }
+                )*
+            }
+        }
+
+        display_field! {
+            self, s:
+            color,
+            rhythm,
+            staminaLeft,
+            staminaRight,
+            aim,
+            aimNoSliders,
+            speed,
+            flashlight,
+            strains,
+            movement,
+        }
+
+        s.finish()
+    }
+}
+
 impl Display for CalculateResult {
     fn fmt(&self, f: &mut Formatter<'_>) -> FmtResult {
         let mut s = f.debug_struct("CalculateResult");
@@ -651,6 +771,7 @@ impl Display for ScoreParams {
         write!(
             f,
             "ScoreParams {{ \
+            mode: {}, \
             mods: {}, \
             n300: {}, \
             n100: {}, \
@@ -663,6 +784,10 @@ impl Display for ScoreParams {
             passedObjects: {}, \
             clockRate: {} \
         }}",
+            match self.mode {
+                Some(ref mode) => mode as &dyn Display,
+                None => &"None" as &dyn Display,
+            },
             self.mods,
             match self.n300 {
                 Some(ref n300) => n300 as &dyn Display,
@@ -710,10 +835,10 @@ impl Display for ScoreParams {
 
 #[pymodule]
 fn rosu_pp_py(_py: Python, m: &PyModule) -> PyResult<()> {
-    m.add_class::<GameMode>()?;
     m.add_class::<ScoreParams>()?;
     m.add_class::<Calculator>()?;
     m.add_class::<CalculateResult>()?;
+    m.add_class::<Strains>()?;
 
     Ok(())
 }
@@ -725,6 +850,7 @@ struct AttributeKey {
 }
 
 impl Hash for AttributeKey {
+    #[inline]
     fn hash<H: Hasher>(&self, state: &mut H) {
         self.mods.hash(state);
         self.passed_objects.hash(state);
@@ -733,6 +859,7 @@ impl Hash for AttributeKey {
 }
 
 impl PartialEq for AttributeKey {
+    #[inline]
     fn eq(&self, other: &Self) -> bool {
         self.mods == other.mods
             && self.passed_objects == other.passed_objects
